@@ -6,10 +6,11 @@ use lazy_static::lazy_static;
 use log::{debug, error};
 use os_pipe::{self, PipeReader, PipeWriter};
 use std::collections::HashMap;
+use std::ffi::{OsStr, OsString};
 use std::fmt;
 use std::fs::{File, OpenOptions};
 use std::io::{Error, ErrorKind, Read, Result, Write};
-use std::path::Path;
+use std::path::{Path, PathBuf};
 use std::process::{Command, Stdio};
 use std::sync::Mutex;
 
@@ -20,7 +21,7 @@ pub struct CmdEnv {
     stderr: CmdOut,
     args: Vec<String>,
     vars: HashMap<String, String>,
-    current_dir: String,
+    current_dir: PathBuf,
 }
 impl CmdEnv {
     /// Returns the arguments for this command
@@ -34,7 +35,7 @@ impl CmdEnv {
     }
 
     /// Returns the current working directory for this command
-    pub fn current_dir(&self) -> &str {
+    pub fn current_dir(&self) -> &Path {
         &self.current_dir
     }
 
@@ -57,9 +58,9 @@ impl CmdEnv {
 type FnFun = fn(&mut CmdEnv) -> CmdResult;
 
 lazy_static! {
-    static ref CMD_MAP: Mutex<HashMap<String, FnFun>> = {
+    static ref CMD_MAP: Mutex<HashMap<OsString, FnFun>> = {
         // needs explicit type, or it won't compile
-        let mut m: HashMap<String, FnFun> = HashMap::new();
+        let mut m: HashMap<OsString, FnFun> = HashMap::new();
         m.insert("".into(), builtin_true);
         Mutex::new(m)
     };
@@ -67,7 +68,7 @@ lazy_static! {
 
 #[doc(hidden)]
 pub fn export_cmd(cmd: &'static str, func: FnFun) {
-    CMD_MAP.lock().unwrap().insert(cmd.to_string(), func);
+    CMD_MAP.lock().unwrap().insert(OsString::from(cmd), func);
 }
 
 /// set debug mode or not, false by default
@@ -88,7 +89,7 @@ pub fn set_pipefail(enable: bool) {
 #[derive(Default)]
 pub struct GroupCmds {
     group_cmds: Vec<(Cmds, Option<Cmds>)>, // (cmd, orCmd) pairs
-    current_dir: String,
+    current_dir: PathBuf,
 }
 
 impl GroupCmds {
@@ -169,7 +170,7 @@ impl Cmds {
         &self.full_cmds
     }
 
-    fn spawn(&mut self, current_dir: &mut String) -> Result<CmdChildren> {
+    fn spawn(&mut self, current_dir: &mut PathBuf) -> Result<CmdChildren> {
         if std::env::var("CMD_LIB_DEBUG") == Ok("1".into()) {
             debug!("Running {} ...", self.get_full_cmds());
         }
@@ -195,41 +196,41 @@ impl Cmds {
         Ok(CmdChildren::from(children))
     }
 
-    fn run_cmd(&mut self, current_dir: &mut String) -> CmdResult {
+    fn run_cmd(&mut self, current_dir: &mut PathBuf) -> CmdResult {
         self.spawn(current_dir)?.wait_cmd_result_nolog()
     }
 
-    fn run_fun(&mut self, current_dir: &mut String) -> FunResult {
+    fn run_fun(&mut self, current_dir: &mut PathBuf) -> FunResult {
         self.spawn(current_dir)?.wait_fun_result_nolog()
     }
 }
 
 #[doc(hidden)]
 pub enum Redirect {
-    FileToStdin(String),
+    FileToStdin(PathBuf),
     StdoutToStderr,
     StderrToStdout,
-    StdoutToFile(String, bool),
-    StderrToFile(String, bool),
+    StdoutToFile(PathBuf, bool),
+    StderrToFile(PathBuf, bool),
 }
 impl fmt::Debug for Redirect {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
-            Redirect::FileToStdin(path) => f.write_str(&format!("< {}", path)),
+            Redirect::FileToStdin(path) => f.write_str(&format!("< {}", path.display())),
             Redirect::StdoutToStderr => f.write_str(">&2"),
             Redirect::StderrToStdout => f.write_str("2>&1"),
             Redirect::StdoutToFile(path, append) => {
                 if *append {
-                    f.write_str(&format!("1>> {}", path))
+                    f.write_str(&format!("1>> {}", path.display()))
                 } else {
-                    f.write_str(&format!("1> {}", path))
+                    f.write_str(&format!("1> {}", path.display()))
                 }
             }
             Redirect::StderrToFile(path, append) => {
                 if *append {
-                    f.write_str(&format!("2>> {}", path))
+                    f.write_str(&format!("2>> {}", path.display()))
                 } else {
-                    f.write_str(&format!("2> {}", path))
+                    f.write_str(&format!("2> {}", path.display()))
                 }
             }
         }
@@ -240,7 +241,7 @@ impl fmt::Debug for Redirect {
 pub struct Cmd {
     // for parsing
     in_cmd_map: bool,
-    args: Vec<String>,
+    args: Vec<OsString>,
     vars: HashMap<String, String>,
     redirects: Vec<Redirect>,
 
@@ -271,20 +272,21 @@ impl Default for Cmd {
 }
 
 impl Cmd {
-    pub fn add_arg(mut self, arg: String) -> Self {
+    pub fn add_arg(mut self, arg: OsString) -> Self {
+        let arg_str = arg.to_string_lossy().to_string();
         if self.args.is_empty() {
-            let v: Vec<&str> = arg.split('=').collect();
+            let v: Vec<&str> = arg_str.split('=').collect();
             if v.len() == 2 && v[0].chars().all(|c| c.is_ascii_alphanumeric() || c == '_') {
-                self.vars.insert(v[0].to_owned(), v[1].to_owned());
+                self.vars.insert(v[0].into(), v[1].into());
                 return self;
             }
-            self.in_cmd_map = CMD_MAP.lock().unwrap().contains_key(arg.as_str());
+            self.in_cmd_map = CMD_MAP.lock().unwrap().contains_key(&arg);
         }
         self.args.push(arg);
         self
     }
 
-    pub fn add_args(mut self, args: Vec<String>) -> Self {
+    pub fn add_args(mut self, args: Vec<OsString>) -> Self {
         for arg in args {
             self = self.add_arg(arg);
         }
@@ -296,7 +298,7 @@ impl Cmd {
         self
     }
 
-    fn arg0(&self) -> String {
+    fn arg0(&self) -> OsString {
         if self.args.is_empty() {
             "".into()
         } else {
@@ -324,7 +326,7 @@ impl Cmd {
 
     fn gen_command(mut self) -> Self {
         if !self.in_cmd_map {
-            let cmds: Vec<String> = self.args.to_vec();
+            let cmds: Vec<OsString> = self.args.to_vec();
             let mut cmd = Command::new(&cmds[0]);
             cmd.args(&cmds[1..]);
             for (k, v) in self.vars.iter() {
@@ -335,7 +337,7 @@ impl Cmd {
         self
     }
 
-    fn spawn(mut self, current_dir: &mut String) -> Result<CmdChild> {
+    fn spawn(mut self, current_dir: &mut PathBuf) -> Result<CmdChild> {
         let arg0 = self.arg0();
         let full_cmd = self.debug_str();
         if arg0 == "cd" {
@@ -348,9 +350,17 @@ impl Cmd {
         } else if self.in_cmd_map {
             let pipe_out = self.stdout_logging.is_none();
             let mut env = CmdEnv {
-                args: self.args,
+                args: self
+                    .args
+                    .into_iter()
+                    .map(|s| s.to_string_lossy().to_string())
+                    .collect(),
                 vars: self.vars,
-                current_dir: current_dir.clone(),
+                current_dir: if current_dir.as_os_str().is_empty() {
+                    std::env::current_dir()?
+                } else {
+                    current_dir.clone()
+                },
                 stdin: if let Some(redirect_in) = self.stdin_redirect.take() {
                     redirect_in
                 } else {
@@ -390,7 +400,7 @@ impl Cmd {
             let mut cmd = self.std_cmd.take().unwrap();
 
             // setup current_dir
-            if !current_dir.is_empty() {
+            if !current_dir.as_os_str().is_empty() {
                 cmd.current_dir(current_dir.clone());
             }
 
@@ -421,7 +431,7 @@ impl Cmd {
         }
     }
 
-    fn run_cd_cmd(&self, current_dir: &mut String) -> CmdResult {
+    fn run_cd_cmd(&self, current_dir: &mut PathBuf) -> CmdResult {
         if self.args.len() == 1 {
             return Err(Error::new(ErrorKind::Other, "cd: missing directory"));
         } else if self.args.len() > 2 {
@@ -429,23 +439,23 @@ impl Cmd {
             return Err(Error::new(ErrorKind::Other, err_msg));
         }
 
-        let dir = &self.args[1];
-        if !std::path::Path::new(&dir).is_dir() {
-            let err_msg = format!("cd: {}: No such file or directory", dir);
+        let dir = PathBuf::from(&self.args[1]);
+        if !dir.is_dir() {
+            let err_msg = format!("cd {}: No such file or directory", dir.display());
             error!("{}", err_msg);
             return Err(Error::new(ErrorKind::Other, err_msg));
         }
 
-        if let Err(e) = Path::new(dir).access(AccessMode::EXECUTE) {
-            error!("cd {}: {}", dir, e);
+        if let Err(e) = dir.access(AccessMode::EXECUTE) {
+            error!("cd {}: {}", dir.display(), e);
             return Err(e);
         }
 
-        *current_dir = dir.clone();
+        *current_dir = dir;
         Ok(())
     }
 
-    fn open_file(path: &str, read_only: bool, append: bool) -> Result<File> {
+    fn open_file(path: &Path, read_only: bool, append: bool) -> Result<File> {
         if read_only {
             OpenOptions::new().read(true).open(path)
         } else {
@@ -484,7 +494,7 @@ impl Cmd {
         for redirect in self.redirects.iter() {
             match redirect {
                 Redirect::FileToStdin(path) => {
-                    self.stdin_redirect = Some(if path == "/dev/null" {
+                    self.stdin_redirect = Some(if path == Path::new("/dev/null") {
                         CmdIn::CmdNull
                     } else {
                         CmdIn::CmdFile(Self::open_file(path, true, false)?)
@@ -505,14 +515,14 @@ impl Cmd {
                     }
                 }
                 Redirect::StdoutToFile(path, append) => {
-                    self.stdout_redirect = Some(if path == "/dev/null" {
+                    self.stdout_redirect = Some(if path == Path::new("/dev/null") {
                         CmdOut::CmdNull
                     } else {
                         CmdOut::CmdFile(Self::open_file(path, false, *append)?)
                     });
                 }
                 Redirect::StderrToFile(path, append) => {
-                    self.stderr_redirect = Some(if path == "/dev/null" {
+                    self.stderr_redirect = Some(if path == Path::new("/dev/null") {
                         CmdOut::CmdNull
                     } else {
                         CmdOut::CmdFile(Self::open_file(path, false, *append)?)
@@ -524,13 +534,54 @@ impl Cmd {
     }
 }
 
+#[doc(hidden)]
+pub trait AsOsStr {
+    fn as_os_str(&self) -> OsString;
+}
+
+impl<T: ToString> AsOsStr for T {
+    fn as_os_str(&self) -> OsString {
+        self.to_string().into()
+    }
+}
+
+#[doc(hidden)]
+#[derive(Default, Debug)]
+pub struct CmdString(OsString);
+impl CmdString {
+    pub fn append<T: AsRef<OsStr>>(mut self, value: T) -> Self {
+        self.0.push(value);
+        self
+    }
+
+    pub fn into_os_string(self) -> OsString {
+        self.0
+    }
+
+    pub fn into_path_buf(self) -> PathBuf {
+        self.0.into()
+    }
+}
+
+impl<T: ?Sized + AsRef<OsStr>> From<&T> for CmdString {
+    fn from(s: &T) -> Self {
+        Self(s.as_ref().into())
+    }
+}
+
+impl fmt::Display for CmdString {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_str(&self.0.to_string_lossy())
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
 
     #[test]
     fn test_run_piped_cmds() {
-        let mut current_dir = String::new();
+        let mut current_dir = PathBuf::new();
         assert!(Cmds::default()
             .pipe(Cmd::default().add_args(vec!["echo".into(), "rust".into()]))
             .pipe(Cmd::default().add_args(vec!["wc".into()]))
@@ -540,7 +591,7 @@ mod tests {
 
     #[test]
     fn test_run_piped_funs() {
-        let mut current_dir = String::new();
+        let mut current_dir = PathBuf::new();
         assert_eq!(
             Cmds::default()
                 .pipe(Cmd::default().add_args(vec!["echo".into(), "rust".into()]))
@@ -562,10 +613,10 @@ mod tests {
 
     #[test]
     fn test_stdout_redirect() {
-        let mut current_dir = String::new();
+        let mut current_dir = PathBuf::new();
         let tmp_file = "/tmp/file_echo_rust";
         let mut write_cmd = Cmd::default().add_args(vec!["echo".into(), "rust".into()]);
-        write_cmd = write_cmd.add_redirect(Redirect::StdoutToFile(tmp_file.to_string(), false));
+        write_cmd = write_cmd.add_redirect(Redirect::StdoutToFile(PathBuf::from(tmp_file), false));
         assert!(Cmds::default()
             .pipe(write_cmd)
             .run_cmd(&mut current_dir)
